import codecs
import copy
import glob
import os
import re
import shutil
import subprocess
import sys

try:
    import polib
except ImportError:
    print('Missing prerequisites. Please run: '
          '"sudo apt install python3-polib"')
    sys.exit(1)

from pelicanconf import PATH

# This defines how complete we expect translations to be before we
# generate HTML from them. Needs to be string.
TRANSLATION_COMPLETION_PERCENTAGE = '0'

BCP47_OVERRIDES = (
    ('zh_CN', 'zh-hans'),
    ('zh_TW', 'zh-hant'),
)


class PO4A(object):
    def __init__(self, pot_file):
        self.default_args = [
            '-f', 'text',
            '-o', 'markdown',
            '-M', 'utf-8',
            ]
        self.pot_file = pot_file
        if not shutil.which('po4a'):
            print('Missing prerequisites. Please run: "sudo apt install po4a"')
            sys.exit(1)

    def run(self, po4a_command, additional_args, with_output=False):
        args = copy.copy(self.default_args)
        args += additional_args
        if with_output:
            ret = subprocess.Popen([po4a_command]+args, stdout=subprocess.PIPE)
        else:
            ret = subprocess.call([po4a_command]+args)
        return ret

    def gettextize(self, documents):
        args = []
        for document in documents:
            args += ['-m', document]
        args += [
            '-p', self.pot_file,
            '-L', 'utf-8',
            ]
        return self.run('po4a-gettextize', args)

    def updatepo(self, langs, documents):
        for lang in langs:
            args = []
            for document in documents:
                args += ['-m', document]
            args += ['-p', lang.file_name]
            ret = self.run('po4a-updatepo', args)
            if ret:
                return False
        return True

    def translate(self, doc, lang):
        args = [
            '-k', TRANSLATION_COMPLETION_PERCENTAGE,
            '-m', doc,
            '-p', lang.file_name,
            '-L', 'utf-8',
            ]
        return self.run('po4a-translate', args, with_output=True)


class Documents(object):
    def __init__(self):
        self.docs = []
        for dirpath, dirnames, filenames in os.walk(PATH):
            for filename in filenames:
                self.docs += [os.path.join(dirpath, filename)]

    def translated_doc(self, file_name, lang):
        match = [doc for doc in self.docs
                 if os.path.basename(doc) == os.path.basename(file_name)]
        if not match:
            return None
        return '%s.%s.md' % (match[0].split('.md')[0],
                             lang.bcp47_code)

    def _call_po4a_translate(self, doc, lang, po4a):
        res = po4a.translate(doc, lang)
        output = codecs.decode(res.communicate()[0])
        broken_title_line = [line for line in output.split('\n')
                             if line.lower().startswith('title:')][0]
        rest = [line for line in output.split('\n')
                if not line.lower().startswith('title')]
        output = '\n'.join(rest)
        return (broken_title_line, output)

    def write_translated_markdown(self, lang, po4a):
        for doc in self.docs:
            (broken_title_line, output) = \
                self._call_po4a_translate(doc, lang, po4a)
            new_path = self.translated_doc(doc, lang)
            text = "%s\nDate:\n\n" % (broken_title_line)
            text += output
            if os.path.exists(new_path):
                os.remove(new_path)
            if not os.path.exists(os.path.dirname(new_path)):
                os.makedirs(os.path.dirname(new_path))
            with open(new_path, 'w', encoding='utf-8') as f:
                f.write(text)


class Language(object):
    def __init__(self, po_file, documents):
        self.file_name = po_file
        self.gettext_code = os.path.basename(po_file).split('.po')[0]
        self.bcp47_code = self._find_bcp47_code()
        self.documents = documents

    def _find_bcp47_code(self):
        if self.gettext_code not in [c[0] for c in BCP47_OVERRIDES]:
            return self.gettext_code.lower().replace('_', '-')
        return [c[1] for c in BCP47_OVERRIDES
                if c[0] == self.gettext_code][0]

    def rewrite_links(self):
        po_file = polib.pofile(self.file_name)
        link_regex = r'\[.+?\]\(\{filename\}(.+?)\)'
        for entry_group in [po_file.translated_entries(),
                            po_file.fuzzy_entries(),
                            po_file.untranslated_entries()]:
            for entry in entry_group:
                if '{filename}' in entry.msgid:
                    link_msgid = re.findall(link_regex, entry.msgid)[0]
                    link_msgstr = list(re.findall(link_regex, entry.msgstr))
                    translated_doc = os.path.basename(
                        self.documents.translated_doc(
                            link_msgid, self))
                    if not link_msgstr:
                        entry.msgstr = entry.msgid
                        link_msgstr = [link_msgid]
                    entry.msgstr = entry.msgstr.replace(link_msgstr[0],
                                                        translated_doc)
        po_file.save(self.file_name)


class Translations(object):
    def __init__(self):
        self._cleanup()
        self.translations_dir = os.path.abspath(os.path.join(PATH, '../po'))
        self.available_languages = []
        self.documents = Documents()
        for po_filename in glob.glob(self.translations_dir+'/*.po'):
            self.available_languages += [Language(po_filename, self.documents)]
        self.fake_lang_code = 'en_US'
        self.fake_po_file = os.path.join(self.translations_dir,
                                         '%s.po' % self.fake_lang_code)
        self.pot_file = os.path.join(self.translations_dir,
                                     "help.pot")
        self.po4a = PO4A(self.pot_file)

    def __del__(self):
        if os.path.exists(self.fake_po_file):
            os.remove(self.fake_po_file)

    def _cleanup(self):
        r = subprocess.Popen(['bzr', 'ignored'], stdout=subprocess.PIPE)
        files = [os.path.join(PATH, '../..', f.split(' ')[0])
                 for f in codecs.decode(r.communicate()[0]).split('\n')
                 if f.strip() != '']
        files = [f for f in files if os.path.exists(f)]
        for f in files:
            try:
                shutil.rmtree(f)
            except NotADirectoryError:
                os.remove(f)

    def generate_pot_file(self):
        return self.po4a.gettextize(self.documents.docs)

    def update_po_files(self):
        return self.po4a.updatepo(self.available_languages,
                                  self.documents.docs)

    def rewrite_links(self):
        self._generate_fake_pofile()
        for lang in self.available_languages:
            lang.rewrite_links()

    # we generate a fake translation for en-US which is going to be
    # the default
    def _generate_fake_pofile(self):
        if os.path.exists(self.fake_po_file):
            os.remove(self.fake_po_file)
        shutil.copy(self.pot_file, self.fake_po_file)
        self.available_languages += [Language(self.fake_po_file,
                                              self.documents)]

    def generate_translations(self):
        for lang in self.available_languages:
            self.documents.write_translated_markdown(lang, self.po4a)
